/// <reference path="../native.ts" />
/// <reference path="../object.ts" />
/* tslint:disable:no-bitwise */

namespace trusted.cms {

    const DEFAULT_DATA_FORMAT: DataFormat = DataFormat.DER;

    export enum SignedDataContentType {
        url,
        buffer,
        hash
    }

    export interface ISignedDataContent {
        type: SignedDataContentType;
        data: string | Buffer | IHashData;
    }
    export interface IHashData {
        value: Buffer;
        alg_id: trusted.HashAlg
    }

    /**
     * Signed data policy
     *
     * @enum {number}
     */
    enum SignedDataPolicy {
        // text = 0x1,
        // noCertificates = 0x2,
        // noContentVerify = 0x4,
        // noAttributeVerify = 0x8,
        // noSignatures = noAttributeVerify | noContentVerify,
        // noIntern = 0x10,
        // noSignerCertificateVerify = 0x20,
        // noVerify = 0x20,
        detached = 0x4,
        // binary = 0x80,
        noAttributes = 0x200,
        // noSmimeCap = 0x200,
        // noOldMimeType = 0x400,
        // crlFEOL = 0x800,
        // stream = 0x1000,
        // noCrtl = 0x2000,
        // partial = 0x4000,
        // reuseDigest = 0x8000,
        // useKeyId = 0x10000,
        // debugDecrypt = 0x20000,
    }

    /**
     * Get name
     *
     * @param {*} e
     * @param {string} name
     * @returns {*}
     */
    function EnumGetName(e: any, name: string): any {
        "use strict";

        for (const i in e) {
            if (i.toString().toLowerCase() === name.toLowerCase()) {
                return { name: i, value: e[i] };
            }
        }
        return undefined;
    }

    export enum StampType {
        stContent = 1,
        stSignature = 2,
        stEscStamp = 4
    }

    export class TimestampParams extends BaseObject<native.CMS.TimestampParams> {
        /**
         * Creates an instance of TimestampParams.
         *
         *
         * @memberOf TimestampParams
         */
        constructor() {
            super();

            this.handle = new native.CMS.TimestampParams();
        }

        /**
         * Return time stamp type
         *
         * @type {StampType}
         * @memberOf TimestampParams
         */
        get stampType(): number {
            return this.handle.getStampType();
        }

        /**
         * Set time stamp type
         *
         *
         * @memberOf TimestampParams
         */
        set stampType(stmp: number) {
            this.handle.setStampType(stmp);
        }

        /**
         * Return connection settings for time stamp service
         *
         * @type {trusted.utils.ConnectionSettings}
         * @memberOf TimestampParams
         */
        get connSettings(): trusted.utils.ConnectionSettings {
            var connSett = new trusted.utils.ConnectionSettings();
            connSett.handle = this.handle.getConnSettings();
            return connSett;
        }

        /**
         * Set connection settings for time stamp service
         *
         *
         * @memberOf TimestampParams
         */
        set connSettings(connSett: trusted.utils.ConnectionSettings) {
            this.handle.setConnSettings(connSett.handle);
        }

        /**
         * Return time stamp hash algorithm OID
         *
         * @type {String}
         * @memberOf TimestampParams
         */
        get tspHashAlg(): String {
            return this.handle.getTspHashAlg();
        }

        /**
         * Set time stamp hash algorithm OID
         *
         *
         * @memberOf TimestampParams
         */
        set tspHashAlg(hashAlg: String) {
            this.handle.setTspHashAlg(hashAlg);
        }
    }

    /**
    * Supported CAdES types
    *
    * @enum {number}
    */
    export enum CadesType {
        ctCadesXLT1 = 1,
        ctCadesT = 2,
        //stCadesA = 3
    }

    export class CadesParams extends BaseObject<native.CMS.CadesParams> {
        /**
         * Creates an instance of TimestampParams.
         *
         *
         * @memberOf CadesParams
         */
        constructor() {
            super();
            this.handle = new native.CMS.CadesParams();
        }

        /**
         * Return CAdES type
         *
         * @type {CadesType}
         * @memberOf CadesParams
         */
        get cadesType(): CadesType {
            return this.handle.getCadesType();
        }

        /**
         * Set CAdES type
         *
         *
         * @memberOf CadesParams
         */
        set cadesType(signType: CadesType) {
            this.handle.setCadesType(signType);
        }

        /**
         * Return connection settings for time stamp service
         *
         * @type {trusted.utils.ConnectionSettings}
         * @memberOf CadesParams
         */
        get connSettings(): trusted.utils.ConnectionSettings {
            var connSett = new trusted.utils.ConnectionSettings();
            connSett.handle = this.handle.getConnSettings();
            return connSett;
        }

        /**
         * Set connection settings for time stamp service
         *
         *
         * @memberOf CadesParams
         */
        set connSettings(connSett: trusted.utils.ConnectionSettings) {
            this.handle.setConnSettings(connSett.handle);
        }

        /**
         * Return time stamp hash algorithm OID
         *
         * @type {String}
         * @memberOf CadesParams
         */
        get tspHashAlg(): String {
            return this.handle.getTspHashAlg();
        }

        /**
         * Set time stamp hash algorithm OID
         *
         *
         * @memberOf CadesParams
         */
        set tspHashAlg(hashAlg: String) {
            this.handle.setTspHashAlg(hashAlg);
        }

        /**
         * Return connection settings for OCSP service
         *
         * @type {trusted.utils.ConnectionSettings}
         * @memberOf CadesParams
         */
        get ocspSettings(): trusted.utils.ConnectionSettings {
            var connSett = new trusted.utils.ConnectionSettings();
            connSett.handle = this.handle.getOcspSettings();
            return connSett;
        }

        /**
         * Set connection settings for time stamp service
         *
         *
         * @memberOf CadesParams
         */
        set ocspSettings(connSett: trusted.utils.ConnectionSettings) {
            this.handle.setOcspSettings(connSett.handle);
        }

    }

    /**
     * Wrap CMS_ContentInfo
     *
     * @export
     * @class SignedData
     * @extends {BaseObject<native.CMS.SignedData>}
     */
    export class SignedData extends BaseObject<native.CMS.SignedData> {
        /**
         * Load signed data from file location
         *
         * @static
         * @param {string} filename File location
         * @param {DataFormat} [format] PEM | DER
         * @returns {SignedData}
         *
         * @memberOf SignedData
         */
        public static load(filename: string, format?: DataFormat): SignedData {
            const cms: SignedData = new SignedData();
            cms.handle.load(filename, format);
            return cms;
        }

        /**
         * Load signed data from memory
         *
         * @static
         * @param {Buffer} buffer
         * @param {DataFormat} [format=DEFAULT_DATA_FORMAT]
         * @returns {SignedData}
         *
         * @memberOf SignedData
         */
        public static import(buffer: Buffer, format: DataFormat = DEFAULT_DATA_FORMAT): SignedData {
            const cms: SignedData = new SignedData();
            cms.handle.import(buffer, format);
            return cms;
        }

        private prContent: ISignedDataContent = undefined;

        /**
         * Creates an instance of SignedData.
         *
         *
         * @memberOf SignedData
         */
        constructor() {
            super();

            this.handle = new native.CMS.SignedData();
        }

        /**
         * Return content of signed data
         *
         * @type {ISignedDataContent}
         * @memberOf SignedData
         */
        get content(): ISignedDataContent {
            if (!this.prContent && !this.isDetached()) {
                // Извлечь содержимое из подписи
                const buf: Buffer = this.handle.getContent();
                this.prContent = {
                    data: buf,
                    type: SignedDataContentType.buffer,
                };
            }
            return this.prContent;
        }

        /**
         * Check if content is available
         *
         * @type {boolean}
         * @memberOf SignedData
         */
        get isContentAvailable(): boolean {
            return this.handle.getIsContentAvailable();
        }

        /**
         * Write content of signed data into file asynchronously
         *
         * @returns {Promise<void>}
         * @memberOf SignedData
         */
        public writeContent(targetFileName: string, done: (error?: string) => void) {
            if (!this.prContent && !this.isDetached()) {
                this.handle.writeContent(targetFileName, done);
                return;
            }

            done("Content is detached");
        }

        /**
         * Set content v to signed data
         *
         *
         * @memberOf SignedData
         */
        set content(v: ISignedDataContent) {
            let data: any;
            if (v.type === SignedDataContentType.url) {
                data = v.data.toString();
            } else if (v.type === SignedDataContentType.buffer) {
                data = Buffer.from(v.data as any);
            }

            if (v.type === SignedDataContentType.hash) {
                let hash = v.data as IHashData;
                if (!hash.value) {
                    throw new TypeError("SignedData::content: hash value is not specified for 'hash' content type");
                }
                this.handle.setContentAsHash(hash.value, hash.alg_id);
            } else {
                this.handle.setContent(data);
            }
            this.prContent = v;
        }

        /**
        * Return sign policys
        *
        * @type {Array<string>}
        * @memberOf SignedData
        */
        get policies(): string[] {
            const p: string[] = new Array<string>();

            const flags: number = this.handle.getFlags();

            for (const i in SignedDataPolicy) {
                if (+i & flags) {
                    p.push(SignedDataPolicy[i]);
                }
            }

            return p;
        }

        /**
        * Set sign policies
        *
        *
        * @memberOf SignedData
        */
        set policies(v: string[]) {
            let flags: number = 0;
            for (const item of v) {
                const flag: any = EnumGetName(SignedDataPolicy, item);
                if (flag) {
                    flags |= +flag.value;
                }
            }

            this.handle.setFlags(flags);
        }

        /**
         *  Free signed content
         *
         * @returns {void}
         * @memberof SignedData
         */
        public freeContent(): void {
            return this.handle.freeContent();
        }

        /**
         * Return true if sign detached
         *
         * @returns {boolean}
         *
         * @memberOf SignedData
         */
        public isDetached(): boolean {
            return this.handle.isDetached();
        }

        /**
         * Return certificates collection or certificate by index (if request)
         *
         * @param {number} [index]
         * @returns {*}
         *
         * @memberOf SignedData
         */
        public certificates(index?: number): any {
            const certs: pki.CertificateCollection = new pki.CertificateCollection(this.handle.getCertificates());
            if (index !== undefined) {
                return certs.items(index);
            }
            return certs;
        }

        ///**
        // * Return certificates collection
        // *
        // * @returns {CertificateCollection}
        // *
        // * @memberOf SignedData
        // */
        //public certificates(): pki.CertificateCollection;

        ///**
        // * Return certificates collection or certificate by index (if request)
        // *
        // * @param {number} [index]
        // * @returns {*}
        // *
        // * @memberOf SignedData
        // */
        //public certificates(index?: number): any {
        //    const certs: pki.CertificateCollection = new pki.CertificateCollection(this.handle.getCertificates());
        //    if (index !== undefined) {
        //        return certs.items(index);
        //    }
        //    return certs;
        //}

        /**
        * Return signer by index
        *
        * @param {number} index
        * @returns {Signer}
        *
        * @memberOf SignedData
        */
        public signers(index: number): Signer;

        /**
        * Return signers collection
        *
        * @returns {SignerCollection}
        *
        * @memberOf SignedData
        */
        public signers(): SignerCollection;

        /**
        * Return signers collection or signer by index (if request)
        *
        * @param {number} [index]
        * @returns {*}
        *
        * @memberOf SignedData
        */
        public signers(index?: number): any {
            const signers: SignerCollection = new SignerCollection(this.handle.getSigners());
            if (index !== undefined) {
                return signers.items(index);
            }
            return signers;
        }

        /**
         * Load sign from file location
         *
         * @param {string} filename File location
         * @param {DataFormat} [format] PEM | DER
         *
         * @memberOf SignedData
         */
        public load(filename: string, format?: DataFormat): void {
            this.handle.load(filename, format);
        }

        /**
         * Load sign asynchronously from file location
         *
         * @param {string} filename File location
         * @param {DataFormat} [format] PEM | DER
         * @param {(message: string) => void} done Done callback
         *
         * @memberOf SignedData
         */
        public loadAsync(filename: string, format: DataFormat, done: (message: string) => void): void {
            this.handle.loadAsync(filename, format, done);
        }

        /**
         * Load sign from memory
         *
         * @param {Buffer} buffer
         * @param {DataFormat} [format=DEFAULT_DATA_FORMAT] PEM | DER (default)
         *
         * @memberOf SignedData
         */
        public import(buffer: Buffer, format: DataFormat = DEFAULT_DATA_FORMAT): void {
            this.handle.import(buffer, format);
        }

        /**
         * Load sign asynchronously from memory
         *
         * @param {Buffer} buffer
         * @param {DataFormat} [format=DEFAULT_DATA_FORMAT] PEM | DER (default)
         * @param {(message: string) => void} done Done callback
         *
         * @memberOf SignedData
         */
        public importAsync(buffer: Buffer, format: DataFormat, done: (message: string) => void): void {
            this.handle.importAsync(buffer, format, done);
        }

        /**
         * Save sign to memory
         *
         * @param {DataFormat} [format=DEFAULT_DATA_FORMAT] PEM | DER (default)
         * @returns {Buffer}
         *
         * @memberOf SignedData
         */
        public export(format: DataFormat = DEFAULT_DATA_FORMAT): Buffer {
            return this.handle.export(format);
        }

        /**
         * Save sign to memory asynchronously
         *
         * @param {DataFormat} [format=DEFAULT_DATA_FORMAT] PEM | DER (default)
         * @returns {Buffer}
         * @param {(error: string, result: Buffer) => void} done Callback to get returned value or error
         *
         * @memberOf SignedData
         */
        public exportAsync(
            format: DataFormat = DEFAULT_DATA_FORMAT,
            done: (error: string, result: Buffer) => void): void {
            this.handle.exportAsync(format, done);
        }

        /**
         * Write sign to file
         *
         * @param {string} filename File location
         * @param {DataFormat} [format=DEFAULT_DATA_FORMAT] PEM | DER (default)
         *
         * @memberOf SignedData
         */
        public save(filename: string, format: DataFormat): void {
            this.handle.save(filename, format);
        }

        /**
         * Write sign to file asynchronously
         *
         * @param {string} filename File location
         * @param {DataFormat} format PEM | DER
         * @param {(message: string)=>void} done Callback to process asynchronous result
         *
         * @memberOf SignedData
         */
        public saveAsync(filename: string, format: DataFormat, done: (message: string) => void): void {
            this.handle.saveAsync(filename, format, done);
        }

        ///**
        // * Create new signer
        // *
        // * @param {Certificate} cert Signer certificate
        // * @param {Key} key Private key for signer certificate
        // * @returns {Signer}
        // *
        // * @memberOf SignedData
        // */
        //public createSigner(cert: pki.Certificate, key: pki.Key): Signer {
        //    const signer: any = this.handle.createSigner(cert.handle, key.handle);
        //    return new Signer(signer);
        //}

        /**
         * Verify signature
         *
         * @param {Signer} [signer] Certificate
         * @returns {boolean}
         *
         * @memberOf SignedData
         */
        public verify(signer?: cms.Signer): boolean {
            if (signer) {
                return this.handle.verify(signer.handle);
            } else {
                return this.handle.verify();
            }
        }

        /**
         * Verify signature asynchronously
         *
         * @param {Signer} [signer] Certificate
         * @param {(error: string, result: boolean) => void} done Callback to get returned value or error
         *
         * @memberOf SignedData
         */
        public verifyAsync(
            done: (error: string, result: boolean) => void,
            signer?: cms.Signer): void {
            if (signer) {
                this.handle.verifyAsync(done, signer.handle);
            } else {
                this.handle.verifyAsync(done);
            }
        }

        /**
         * Return signature creation parameters
         *
         * @type {TimestampParams | CadesParams}
         * @memberOf SignedData
         */
        get signParams(): TimestampParams | CadesParams {
            var internalValue = this.handle.getSignParams();
            var result = undefined;

            if (internalValue instanceof native.CMS.TimestampParams) {
                result = new TimestampParams();
                result.handle = internalValue;
            }
            else if (internalValue instanceof native.CMS.CadesParams) {
                result = new CadesParams();
                result.handle = internalValue;
            }

            return result;
        }

        /**
         * Set signature creation parameters
         *
         *
         * @memberOf SignedData
         */
        set signParams(params: TimestampParams | CadesParams) {
            if (params == undefined)
                throw new TypeError("SignedData::signParams: Wrong input param"); //return?

            this.handle.setSignParams(params.handle);
        }

        /**
         * Create sign
         *
         * @param {Certificate} [certs] Certificate
         *
         * @memberOf SignedData
         */
        public sign(cert: pki.Certificate): void {
            this.handle.sign(cert.handle);
        }

        /**
         * Create sign asynchronously
         *
         * @param {Certificate} [certs] Certificate
         * @param {(message: string)=>void} done Callback to process asynchronous result
         *
         * @memberOf SignedData
         */
        public signAsync(cert: pki.Certificate, done: (message: string) => void): void {
            this.handle.signAsync(cert.handle, done);
        }
    }
}
