/// <reference path="../native.ts" />
/// <reference path="../object.ts" />

namespace trusted.pki {

    /**
     * Wrap TSPRequest utility object for hashing data before request sending
     *
     * @export
     * @class TSPRequest
     * @extends {BaseObject<native.PKI.TSPRequest>}
     */
    export class TSPRequest extends BaseObject<native.PKI.TSPRequest> {
        /**
         * Creates an instance of Tsp Request.
         * @param {hashAlgOid: string, dataFileName?: string} [param]
         *
         * @memberOf TSPRequest
         */
        constructor(hashAlgOid: string, dataFileName?: string) {
            super();
            if (dataFileName != undefined) {
                this.handle = new native.PKI.TSPRequest(hashAlgOid, dataFileName);
            } else {
                this.handle = new native.PKI.TSPRequest(hashAlgOid);
            }
        }

        public AddData(data: Buffer): void {
            this.handle.AddData(data);
        }

        public get CertReq(): boolean {
            return this.handle.GetCertReq();
        }

        public set CertReq(certReq: boolean) {
            this.handle.SetCertReq(certReq);
        }

        public get Nonce(): boolean {
            return this.handle.GetNonce();
        }

        public set Nonce(nonce: boolean) {
            this.handle.SetNonce(nonce);
        }

        public get PolicyId(): string {
            return this.handle.GetPolicyId();
        }

        public set PolicyId(policyId: string) {
            this.handle.SetPolicyId(policyId);
        }

        public get HashAlgOid(): string {
            return this.handle.GetHashAlgOid();
        }

        public set HashAlgOid(hashAlgOid: string) {
            this.handle.SetHashAlgOid(hashAlgOid);
        }

        public get DataHash(): Buffer {
            return this.handle.GetDataHash();
        }

        public set DataHash(dataHash: Buffer) {
            this.handle.SetDataHash(dataHash);
        }
    }

    /**
     * Wrap TSP timestamp object and request sending
     *
     * @export
     * @class TSP
     * @extends {BaseObject<native.PKI.TSP>}
     */
    export class TSP extends BaseObject<native.PKI.TSP> {
        /**
         * Creates an instance of Tsp.
         * @param {Buffer, native.UTILS.ConnectionSettings?} [param]
         *
         * @memberOf TSP
         */
        constructor(inData: trusted.pki.TSPRequest | Buffer | native.PKI.TSP, connSettings?: trusted.utils.ConnectionSettings) {
            super();
            if (inData instanceof Buffer) {
                this.handle = new native.PKI.TSP(inData);
            }
            else if ((inData instanceof trusted.pki.TSPRequest)
                && (connSettings instanceof trusted.utils.ConnectionSettings)) {
                this.handle = new native.PKI.TSP(inData.handle, connSettings.handle);
            }
            else if (inData instanceof native.PKI.TSP) {
                this.handle = inData;
            }
            else {
                throw new TypeError("TSP::constructor: Wrong input param");
            }
        }

        public Export(): Buffer {
            return this.handle.Export();
        }

        public get Certificates(): CertificateCollection {
            return CertificateCollection.wrap<native.PKI.CertificateCollection, CertificateCollection>(this.handle.Certificates());
        }

        public get TSACertificate(): Certificate | undefined {
            const nativeCert = this.handle.TSACertificate();
            if (!nativeCert) {
                return undefined;
            }
            return Certificate.wrap<native.PKI.Certificate, Certificate>(nativeCert);
        }

        public Verify(): number {
            return this.handle.Verify();
        }

        public VerifyCertificate(cert: Certificate): number {
            return this.handle.VerifyCertificate( cert.handle );
        }

        public get FailInfo(): number {
            return this.handle.FailInfo();
        }

        public get Status(): number {
            return this.handle.Status();
        }

        public get StatusString(): string {
            return this.handle.StatusString();
        }

        public get DataHashAlgOID(): string {
            return this.handle.DataHashAlgOID();
        }

        public get DataHash(): Buffer {
            return this.handle.DataHash();
        }

        public get PolicyID(): string {
            return this.handle.PolicyID();
        }

        public get SerialNumber(): Buffer {
            return this.handle.SerialNumber();
        }

        public get Time(): Date {
            return new Date(this.handle.Time());
        }

        public get Accuracy(): number {
            return this.handle.Accuracy();
        }

        public get Ordering(): boolean {
            return this.handle.Ordering();
        }

        public get HasNonce(): boolean {
            return this.handle.HasNonce();
        }

        public get TsaName(): string {
            return this.handle.TsaName();
        }

        public get TsaNameBlob(): Buffer {
            return this.handle.TsaNameBlob();
        }
    }
}
